from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import BinaryIO, Optional, Sequence, Type

from directorfile.common import Endianness, EndiannessAwareStream, ParsingError


class Resource(metaclass=ABCMeta):
    def __repr__(self):
        return f'<{type(self).__qualname__} at {hex(id(self))}>'

    def load(self, fp: BinaryIO, position: Optional[int] = None, size: int = 0) -> Resource:
        if position is not None:
            fp.seek(position)

        reader = self.parse_tag(fp)
        data_size = reader.read_ui32()
        if size:
            assert size >= data_size
        return self.parse(reader, data_size)

    def parse(self, reader: EndiannessAwareStream, size: int) -> Resource:
        start = reader.get_current_pos()
        self._parse(reader, size)
        reader.jump(start + size)
        return self

    def parse_tag(self, fp: BinaryIO) -> EndiannessAwareStream:
        tag = fp.read(4).decode('ascii')
        if tag == self.TAG:
            endianness = Endianness.BIG_ENDIAN
        elif tag == self.TAG[::-1]:
            endianness = Endianness.LITTLE_ENDIAN
        else:
            raise ParsingError(f'Expected {self.TAG} tag, got {tag} instead')

        reader = EndiannessAwareStream(fp, endianness)
        return reader

    def save(self, fp: BinaryIO, endianness: Endianness, position: Optional[int] = None) -> int:
        if position is not None:
            fp.seek(position)

        stream = EndiannessAwareStream(fp, endianness)
        stream.skip(8)

        start = stream.get_current_pos()
        self.serialize(stream)
        size = stream.get_current_pos() - start
        stream.jump(start - 8)
        self.serialize_header(stream, size)
        stream.skip(size)

        return size

    def serialize(self, stream: EndiannessAwareStream):
        self._serialize(stream)

    def serialize_header(self, writer: EndiannessAwareStream, size: int):
        writer.write_tag(self.TAG)
        writer.write_ui32(size)

    @abstractmethod
    def _parse(self, reader: EndiannessAwareStream, size: int) -> None:
        pass

    @abstractmethod
    def _serialize(self, writer: EndiannessAwareStream) -> None:
        pass

    @property
    @abstractmethod
    def TAG(self) -> str:
        pass


class ArchiveParser(metaclass=ABCMeta):
    def __init__(self, archive: RIFXArchiveResource, reader: EndiannessAwareStream):
        self.archive = archive
        self._reader = reader

    @abstractmethod
    def parse(self):
        pass

    @property
    @abstractmethod
    def TYPES(self) -> str:
        pass


class ArchiveSerializer(metaclass=ABCMeta):
    def __init__(self, endianness: Endianness, director_version: int):
        self.endianness = endianness
        self.director_version = director_version

    def serialize(self, fp: BinaryIO, archive: RIFXArchiveResource):
        stream = EndiannessAwareStream(fp, self.endianness)
        self._serialize(stream, archive)

    @abstractmethod
    def _serialize(self, stream: EndiannessAwareStream, archive: RIFXArchiveResource):
        pass


class FileResource(Resource, metaclass=ABCMeta):
    filename: str

    def __init__(self, filename: str = ''):
        self.filename = filename

    def __repr__(self):
        if self.filename:
            identifier = f'"{self.filename}"'
        else:
            identifier = f'at {hex(id(self))}'

        return f'<{type(self).__qualname__} {identifier}>'


class RIFXArchiveResource(FileResource):
    TAG = 'RIFX'

    PARSERS: Sequence[Type[ArchiveParser]] = tuple()

    _parser: ArchiveParser

    def _parse(self, reader: EndiannessAwareStream, size: int) -> None:
        tag = reader.read_tag()

        for parser_class in self.PARSERS:
            if tag in parser_class.TYPES:
                parser = parser_class(self, reader)
                break
        else:
            raise ParsingError(f'Could not find parser for a {tag} archive')

        self._parser = parser
        parser.parse()

    def _serialize(self, writer: EndiannessAwareStream) -> None:
        raise NotImplementedError('Not directly serializable')



from directorfile.archive.base import ArchiveParser


class ShockwaveArchiveParser(ArchiveParser):
    TYPES = {'FGDC', 'FGDM'}

    def parse(self):
        raise NotImplementedError("Shockwave archives are not yet supported")



from __future__ import annotations

import os
import zlib
from dataclasses import asdict, dataclass
from enum import IntEnum
from typing import BinaryIO, Dict, List, Optional, Sequence, Tuple, Type

from directorfile.archive.base import FileResource, Resource
from directorfile.archive.director import DirectorArchiveParser, DirectorArchiveResource, DirectorArchiveSerializer, \
    MMapResource, RIFXArchiveResource
from directorfile.common import Endianness, EndiannessAwareStream, ParsingError, calculate_alignment_remainder


class FileType(IntEnum):
    DIRECTOR_MOVIE = 0
    DIRECTOR_CAST = 1
    XTRA = 2


@dataclass
class FileRecord:
    filename: str
    type: FileType
    resource: FileResource


class DictResource(Resource):
    TAG = 'Dict'

    HEADER_SIZE = 0x1c
    ENTRY_WIDTH = 0x08

    mapping: Dict[int, str]

    def __init__(self, mapping: Dict[int, str] = None):
        if not mapping:
            self.mapping = {}
        else:
            self.mapping = dict(mapping)

    def _parse(self, reader: EndiannessAwareStream, size: int):
        values_chunk_offset = reader.read_ui32()
        values_chunk_size = reader.read_ui32()

        header_offset = reader.get_current_pos()
        values_base = header_offset + values_chunk_offset

        assert reader.read_ui32() == 0
        assert reader.read_ui32() == 0

        length = reader.read_ui32()
        allocated_length = reader.read_ui32()
        assert allocated_length >= length

        assert reader.read_ui16() == DictResource.HEADER_SIZE
        assert reader.read_ui16() == DictResource.ENTRY_WIDTH

        assert reader.read_ui32() == 0
        assert reader.read_ui32() == 0

        pairs = []
        for i in range(length):
            value_offset = reader.read_ui32()
            key = reader.read_ui32()
            pairs.append((key, value_offset))

        reader.skip(8 * (allocated_length - length))
        assert reader.get_current_pos() == values_base

        assert reader.read_ui32() == 0
        assert reader.read_ui32() == 0

        values_chunk_used = reader.read_ui32()
        assert values_chunk_used <= values_chunk_size

        assert reader.read_ui32() == values_chunk_size
        assert reader.read_ui32() == 20

        mapping = {}
        for key, value_offset in pairs:
            assert key not in mapping
            reader.jump(values_base + value_offset)
            value = reader.read_string()
            mapping[key] = value
        self.mapping = mapping

    def _serialize(self, writer: EndiannessAwareStream) -> None:
        values_chunk_offset = DictResource.HEADER_SIZE + len(self.mapping) * DictResource.ENTRY_WIDTH

        values_chunk_used = 20
        for value in self.mapping.values():
            values_chunk_used += 4 + len(value)
            values_chunk_used += calculate_alignment_remainder(len(value), 4)
        values_chunk_size = values_chunk_used

        writer.write_ui32(values_chunk_offset)
        writer.write_ui32(values_chunk_used)

        header_base = writer.get_current_pos()

        writer.write_ui32(0)
        writer.write_ui32(0)

        writer.write_ui32(len(self.mapping))
        writer.write_ui32(len(self.mapping))

        writer.write_ui16(DictResource.HEADER_SIZE)
        writer.write_ui16(DictResource.ENTRY_WIDTH)

        writer.write_ui32(0)
        writer.write_ui32(0)

        value_offset = 20
        for key, value in self.mapping.items():
            writer.write_ui32(value_offset)
            writer.write_ui32(key)

            value_offset += 4 + len(value)
            value_offset += calculate_alignment_remainder(len(value), 4)

        values_base = writer.get_current_pos()
        assert values_base - header_base == values_chunk_offset

        writer.write_ui32(0)
        writer.write_ui32(0)
        writer.write_ui32(values_chunk_used)
        writer.write_ui32(values_chunk_size)
        writer.write_ui32(20)

        for key, value in self.mapping.items():
            writer.write_string(value)
            # 32-bit alignment
            writer.skip(calculate_alignment_remainder(len(value), 4))

        assert values_chunk_used == writer.get_current_pos() - values_base


class ListResource(Resource):
    TAG = 'List'

    HEADER_SIZE = 0x14
    ENTRY_WIDTH = 0x08

    members: List[Tuple[int, int]]

    def __init__(self, members: List[Tuple[int, int]] = None):
        if not members:
            self.members = []
        else:
            self.members = list(members)

    def _parse(self, reader: EndiannessAwareStream, size: int):
        reader.skip(8)
        length = reader.read_ui32()
        allocated_length = reader.read_ui32()
        assert allocated_length >= length
        assert reader.read_ui16() == 0x14
        assert reader.read_ui16() == 0x08

        pairs = []
        for i in range(length):
            index = reader.read_ui32()
            value = reader.read_ui32()
            pairs.append((index, value))
        self.members = pairs

    def _serialize(self, writer: EndiannessAwareStream) -> None:
        writer.write_ui32(0)
        writer.write_ui32(0)

        writer.write_ui32(len(self.members))
        writer.write_ui32(len(self.members))

        writer.write_ui16(ListResource.HEADER_SIZE)
        writer.write_ui16(ListResource.ENTRY_WIDTH)

        for index, value in self.members:
            writer.write_ui32(index)
            writer.write_ui32(value)


class BadDResource(DictResource):
    TAG = 'BadD'


class RIFFXtraFileResource(FileResource):
    TAG = 'RIFF'

    HEADER_SIZE = 0x1c

    def _parse(self, reader: EndiannessAwareStream, size: int):
        assert reader.read_tag() == 'Xtra'
        assert reader.read_tag() == 'FILE'

        headered_size = reader.read_ui32()
        header_size = reader.read_ui32()
        assert header_size == RIFFXtraFileResource.HEADER_SIZE

        reader.skip(8)
        uncompressed_size = reader.read_ui32()
        reader.skip(4)
        compressed_size = reader.read_ui32()
        reader.skip(4)

        assert headered_size == compressed_size + header_size

        self.data = zlib.decompress(reader.read_buffer(compressed_size))

        assert len(self.data) == uncompressed_size

    def save(self, fp: BinaryIO, endianness: Endianness, position: Optional[int] = None) -> int:
        return super().save(fp, Endianness.BIG_ENDIAN, position)

    def _serialize(self, writer: EndiannessAwareStream) -> None:
        writer.write_tag('Xtra')
        writer.write_tag('FILE')

        compressed_data = zlib.compress(self.data)

        writer.write_ui32(len(compressed_data) + RIFFXtraFileResource.HEADER_SIZE)
        writer.write_ui32(RIFFXtraFileResource.HEADER_SIZE)

        writer.write_ui32(0)
        writer.write_ui32(0)
        writer.write_ui32(len(self.data))
        writer.write_ui32(0)
        writer.write_ui32(len(compressed_data))
        writer.write_ui32(0)

        writer.write_buffer(compressed_data)


class ApplicationArchiveParser(DirectorArchiveParser):
    TYPES = {'APPL'}

    RESOURCE_CLASSES: Dict[str, Type[Resource]] = {
        cls.TAG: cls for cls in (
            ListResource,
            DictResource,
            BadDResource,
        )
    }

    FILE_RESOURCE_CLASSES: Sequence[Type[FileResource]] = [
        DirectorArchiveResource,
        RIFFXtraFileResource
    ]

    files: List[FileRecord]
    badd: Dict

    def parse(self):
        super().parse()
        entries = self._mmap.entries

        list_entry = entries[3]
        assert list_entry.tag == 'List'
        file_type_list = self._fetch_resource(list_entry)
        assert isinstance(file_type_list, ListResource)

        dict_entry = entries[4]
        assert dict_entry.tag == 'Dict'
        filename_dict = self._fetch_resource(dict_entry)
        assert isinstance(filename_dict, DictResource)

        badd_entry = entries[5]
        assert badd_entry.tag == 'BadD'
        badd_dict = self._fetch_resource(badd_entry)
        assert isinstance(badd_dict, DictResource)

        files = []
        for i, (entry_index, file_type) in enumerate(file_type_list.members):
            entry = entries[entry_index]

            assert entry.tag == 'File'

            filename = filename_dict.mapping[i]

            file_resource = self._fetch_resource(entry)

            assert isinstance(file_resource, FileResource)
            file_resource.filename = filename

            files.append(FileRecord(filename, FileType(file_type), file_resource))

        self.files = files
        self.badd = badd_dict.mapping

    def _reconstruct_resource(self, entry: MMapResource.Entry):
        fp = self._reader.fp

        tag = entry.tag
        position = entry.position
        size = entry.size

        if tag == 'File':
            for resource_class in self.FILE_RESOURCE_CLASSES:
                try:
                    return resource_class().load(fp=fp, position=position, size=size)
                except ParsingError:
                    pass
            else:
                fp.seek(position)
                raise ParsingError(f"Unknown file header: {fp.read(12)}")
        else:
            resource_class = self.RESOURCE_CLASSES.get(tag)
            if resource_class is None:
                raise ParsingError(f"Unknown resource type '{tag}'")
            return resource_class().load(fp=fp, position=position, size=size)


class ApplicationArchiveSerializer(DirectorArchiveSerializer):
    def _serialize(self, stream: EndiannessAwareStream, archive: ApplicationArchiveResource):
        xtras_first_index = 6
        movies_first_index = xtras_first_index + len(archive.xtras)
        casts_first_index = movies_first_index + len(archive.movies)

        files = [
            *(
                (xtras_first_index + index, FileType.XTRA, path, resource)
                for index, (path, resource) in enumerate(archive.xtras.items())
            ),
            *(
                (movies_first_index + index, FileType.DIRECTOR_MOVIE, path, resource)
                for index, (path, resource) in enumerate(archive.movies.items())
            ),
            *(
                (casts_first_index + index, FileType.DIRECTOR_CAST, path, resource)
                for index, (path, resource) in enumerate(archive.casts.items())
            ),
        ]

        file_type_list = ListResource([(index, file_type) for index, file_type, path, resource in files])
        filename_dict = DictResource({i: path for i, (index, file_type, path, resource) in enumerate(files)})
        badd_dict = BadDResource(archive.badd)

        resources = {
            3: file_type_list,
            4: filename_dict,
            5: badd_dict,
            **{index: resource for index, file_type, path, resource in files}
        }

        self._serialize_resources(stream, 'APPL', resources)

    def _generate_entries(self, stream, resources):
        entries = super()._generate_entries(stream, resources)
        for entry in entries:
            if entry.tag in ('RIFF', 'RIFX'):
                entry.tag = 'File'
                entry.size += 8
        return entries


class ApplicationArchiveResource(RIFXArchiveResource):
    PARSERS = [ApplicationArchiveParser]

    _parser: ApplicationArchiveParser
    xtras: Dict[str, RIFFXtraFileResource]
    casts: Dict[str, RIFXArchiveResource]
    movies: Dict[str, RIFXArchiveResource]

    badd: Dict[int, str]
    director_version: int

    def __init__(self, filename: str = ''):
        super().__init__(filename)
        self.xtras = {}
        self.casts = {}
        self.movies = {}

    def _parse(self, reader: EndiannessAwareStream, size: int) -> None:
        super()._parse(reader, size)
        for file_record in self._parser.files:
            files_dict = {
                FileType.XTRA: self.xtras,
                FileType.DIRECTOR_CAST: self.casts,
                FileType.DIRECTOR_MOVIE: self.movies,
            }[file_record.type]

            files_dict[os.path.basename(file_record.filename)] = file_record.resource

        self.badd = self._parser.badd

        self.director_version = self._parser.director_version

    def _serialize(self, writer: EndiannessAwareStream) -> None:
        serializer = ApplicationArchiveSerializer(writer.endianness, self.director_version)
        serializer.serialize(writer.fp, self)



from directorfile.archive.base import RIFXArchiveResource
from directorfile.archive.director import DirectorArchiveResource
from directorfile.archive.application import ApplicationArchiveResource

from directorfile.archive.director import load_director_archive


def _init_parsers():
    from directorfile.archive.base import RIFXArchiveResource
    from directorfile.archive.director import DirectorArchiveParser
    from directorfile.archive.shockwave import ShockwaveArchiveParser

    RIFXArchiveResource.PARSERS = [
        DirectorArchiveParser,
        ShockwaveArchiveParser
    ]


_init_parsers()



from abc import ABCMeta
from enum import StrEnum
from io import SEEK_CUR
from struct import pack, unpack
from typing import BinaryIO


class Endianness(StrEnum):
    BIG_ENDIAN = '>'
    LITTLE_ENDIAN = '<'


class EndiannessAwareStream(metaclass=ABCMeta):
    fp: BinaryIO
    endianness: Endianness

    def __init__(self, fp: BinaryIO, endianness: Endianness):
        self.fp = fp
        self.endianness = endianness

    def jump(self, position):
        self.fp.seek(position)

    def skip(self, bytes_number):
        self.fp.seek(bytes_number, SEEK_CUR)

    def get_current_pos(self) -> int:
        return self.fp.tell()

    def read_ui16(self) -> int:
        (num,) = unpack(self.endianness + "H", self.read_buffer(2))
        return num

    def read_i16(self) -> int:
        (num,) = unpack(self.endianness + "h", self.read_buffer(2))
        return num

    def read_ui32(self) -> int:
        (num,) = unpack(self.endianness + "I", self.read_buffer(4))
        return num

    def read_i32(self) -> int:
        (num,) = unpack(self.endianness + "i", self.read_buffer(4))
        return num

    def read_buffer(self, count) -> bytes:
        data = self.fp.read(count)
        return data

    def read_tag(self) -> str:
        tag = self.read_buffer(4)
        if self.endianness == Endianness.LITTLE_ENDIAN:
            tag = tag[::-1]
        return tag.decode("ascii")

    def read_string(self) -> str:
        length = self.read_ui32()
        return self.read_buffer(length).decode('ascii')

    def write_ui16(self, num: int):
        self.fp.write(pack(self.endianness + "H", num))

    def write_i16(self, num: int):
        self.fp.write(pack(self.endianness + "h", num))

    def write_ui32(self, num: int):
        self.fp.write(pack(self.endianness + "I", num))

    def write_i32(self, num: int):
        self.fp.write(pack(self.endianness + "i", num))

    def write_buffer(self, data: bytes):
        self.fp.write(data)

    def write_tag(self, tag: str):
        if self.endianness == Endianness.LITTLE_ENDIAN:
            tag = tag[::-1]
        self.fp.write(tag.encode("ascii"))

    def write_string(self, string: str):
        self.write_ui32(len(string))
        self.write_buffer(string.encode('ascii'))


class ParsingError(Exception):
    pass


def calculate_alignment_remainder(value, alignment):
    return (alignment - value) % alignment



from __future__ import annotations

from dataclasses import dataclass
from typing import BinaryIO, Dict, List, Tuple

from directorfile.archive.base import ArchiveParser, ArchiveSerializer, RIFXArchiveResource, Resource
from directorfile.common import EndiannessAwareStream, calculate_alignment_remainder

DIRECTOR_VERSIONS = {
    0x404: '3.0',
    0x405: '3.1',
    0x45b: '4.0.0',
    0x45d: '4.0.4',
    0x4c1: '5.0',
    0x4c7: '6.0',
    0x57e: '7.0',
    0x640: '8.0',
    0x708: '8.5',
    0x73a: '8.5.1',
    0x742: '10.0',
    0x744: '10.1',
    0x781: '11.0',
    0x782: '11.5.0.593',
    0x783: '11.5.8.612',
    0x79f: '12',
}


class GenericResource(Resource):
    data: bytes

    @property
    def TAG(self):
        return self._tag

    def __init__(self, tag: str):
        self._tag = tag
        self.data = b''

    def __repr__(self):
        return f'<GenericResource "{self._tag}" ({len(self.data)} bytes) at {hex(id(self))}>'

    def _parse(self, reader: EndiannessAwareStream, size: int) -> None:
        self.data = reader.read_buffer(size)

    def _serialize(self, writer: EndiannessAwareStream) -> None:
        writer.write_buffer(self.data)


class IMapResource(Resource):
    TAG = 'imap'

    mmap_position: int
    director_version: int

    def __init__(self, mmap_position: int = None, director_version: int = None):
        self.mmap_position = mmap_position
        self.director_version = director_version

    def _parse(self, reader: EndiannessAwareStream, size: int) -> None:
        assert reader.read_ui32() == 1
        self.mmap_position = reader.read_ui32()
        self.director_version = reader.read_ui32()
        assert self.director_version in DIRECTOR_VERSIONS, f'Unsupported version code 0x{self.director_version:02x}'
        assert reader.read_i32() == 0

        assert reader.read_i32() == 0
        assert reader.read_i32() == 0

    def _serialize(self, writer: EndiannessAwareStream) -> None:
        writer.write_ui32(1)
        writer.write_ui32(self.mmap_position)
        writer.write_ui32(self.director_version)
        writer.write_ui32(0)

        writer.write_ui32(0)
        writer.write_ui32(0)


class MMapResource(Resource):
    TAG = 'mmap'

    entries: List["MMapResource.Entry"]

    HEADER_SIZE = 0x18
    ENTRY_WIDTH = 0x14

    def __init__(self, entries: List["MMapResource.Entry"] = None):
        if entries:
            self.entries = list(entries)
        else:
            self.entries = []

    def _parse(self, reader: EndiannessAwareStream, size: int):
        header_size = reader.read_ui16()
        assert header_size == MMapResource.HEADER_SIZE
        width = reader.read_ui16()
        assert width == MMapResource.ENTRY_WIDTH

        allocated_length = reader.read_ui32()
        length = reader.read_ui32()
        assert allocated_length >= length

        # TODO: support free amd junk indices
        unk_junk_indices = [reader.read_i32(), reader.read_i32()]
        unk_free_index = reader.read_i32()

        entries = []
        for index in range(length):
            tag = reader.read_tag()
            size = reader.read_ui32()
            position = reader.read_ui32()

            reader.skip(8)

            entries.append(MMapResource.Entry(index=index, tag=tag, position=position, size=size))

        assert all(index == -1 or entries[index].tag == 'junk' for index in unk_junk_indices)
        assert unk_free_index == -1 or entries[unk_free_index].tag == 'free'

        self.entries = entries

    def _serialize(self, writer: EndiannessAwareStream) -> None:
        writer.write_ui16(MMapResource.HEADER_SIZE)
        writer.write_ui16(MMapResource.ENTRY_WIDTH)

        writer.write_ui32(len(self.entries))
        writer.write_ui32(len(self.entries))

        # TODO: support free amd junk indices
        writer.write_i32(-1)
        writer.write_i32(-1)
        writer.write_i32(-1)

        for entry in self.entries:
            writer.write_tag(entry.tag)
            writer.write_ui32(entry.size)
            writer.write_ui32(entry.position)

            # TODO: support type-specific entry fields
            writer.write_ui32(0)
            writer.write_ui32(0)

    @staticmethod
    def calculate_needed_size(length):
        return MMapResource.HEADER_SIZE + length * MMapResource.ENTRY_WIDTH

    @dataclass
    class Entry:
        index: int
        tag: str
        position: int
        size: int

        def __repr__(self):
            return f'<MMap Entry for "{self.tag}" @ 0x{self.position:08x} ({self.size} bytes)>'


class DirectorArchiveParser(ArchiveParser):
    TYPES = {'M!07', 'M!08', 'M!85', 'M!93', 'M!95', 'M!97', 'M*07', 'M*08', 'M*85', 'M*95', 'M*97', 'MC07',
             'MC08', 'MC85', 'MC95', 'MC97', 'MMQ5', 'MV07', 'MV08', 'MV85', 'MV93', 'MV95', 'MV97'}

    _mmap: MMapResource

    director_version: int
    entries: List[Tuple[MMapResource.Entry, Resource]]

    _resources: Dict[Tuple[str, int], Resource]

    def __init__(self, archive: RIFXArchiveResource, reader: EndiannessAwareStream):
        super().__init__(archive, reader)
        self.entries = []
        self._resources = {}

    def _populate_fetched_resource(self, resource: Resource, position: int):
        self._resources[(resource.TAG, position)] = resource

    def _fetch_resource(self, entry: MMapResource.Entry) -> Resource:
        resource = self._resources.get((entry.tag, entry.position))
        if resource is None:
            resource = self._reconstruct_resource(entry)
            self._populate_fetched_resource(resource, entry.position)
        return resource

    def _reconstruct_resource(self, entry: MMapResource.Entry) -> Resource:
        return GenericResource(entry.tag).load(self._reader.fp, entry.position, entry.size)

    def parse(self):
        imap_position = self._reader.get_current_pos()

        imap = IMapResource().load(self._reader.fp, imap_position)
        mmap = MMapResource().load(self._reader.fp, imap.mmap_position)

        assert mmap.entries[0].tag == 'RIFX'
        self._populate_fetched_resource(self.archive, mmap.entries[0].position)

        assert mmap.entries[1].tag == 'imap'
        self._populate_fetched_resource(imap, imap_position)

        assert mmap.entries[2].tag == 'mmap'
        self._populate_fetched_resource(mmap, imap.mmap_position)

        self._mmap = mmap

        self.director_version = imap.director_version
        self.entries = [(entry, self._fetch_resource(entry))
                        for entry in mmap.entries[3:]
                        if entry.tag not in ('free', 'junk')]


class DirectorArchiveSerializer(ArchiveSerializer):
    def _serialize(self, stream: EndiannessAwareStream, archive: DirectorArchiveResource):
        self._serialize_resources(stream, 'MV93', archive.resources)

    def _serialize_resources(self, stream: EndiannessAwareStream, archive_type: str, resources: Dict[int, Resource]):
        archive_position = stream.get_current_pos() - 8
        imap_position = archive_position + 12
        mmap_position = imap_position + 8 + 0x18
        mmap_size = MMapResource.calculate_needed_size(max(resources.keys()) + 1)
        resources_offset = mmap_position + 8 + mmap_size
        stream.jump(resources_offset)

        entries = self._generate_entries(stream, resources)
        archive_size = max(entry.position + entry.size + 8 for entry in entries) - archive_position - 8

        entries = [
                      MMapResource.Entry(index=0, tag='RIFX', position=archive_position, size=archive_size),
                      MMapResource.Entry(index=1, tag='imap', position=imap_position, size=0x18),
                      MMapResource.Entry(index=2, tag='mmap', position=mmap_position, size=mmap_size),
                  ] + entries

        MMapResource(entries).save(stream.fp, stream.endianness, mmap_position)
        IMapResource(mmap_position, self.director_version).save(stream.fp, stream.endianness, imap_position)
        stream.jump(archive_position + 8)
        stream.write_tag(archive_type)
        stream.jump(archive_position + 8 + archive_size)

    def _generate_entries(self, stream: EndiannessAwareStream, resources: Dict[int, Resource]):
        entry_dict = {}
        for index, resource in resources.items():
            if index < 3:
                continue

            # 16-bit alignment (critical for Xtra file loading)
            stream.skip(calculate_alignment_remainder(stream.get_current_pos(), 2))

            position = stream.get_current_pos()
            size = resource.save(stream.fp, stream.endianness)
            entry_dict[index] = MMapResource.Entry(index=index, tag=resource.TAG, position=position, size=size)

        entries = [
            entry_dict.get(index) or MMapResource.Entry(index, 'free', 0, 0)
            for index in range(3, max(entry_dict.keys()) + 1)
        ]
        return entries


class DirectorArchiveResource(RIFXArchiveResource):
    _parser: DirectorArchiveParser

    resources: Dict[int, Resource]
    director_version: int

    def __init__(self, filename: str = '', resources: Dict[int, Resource] = None, director_version: int = None):
        super().__init__(filename=filename)
        if not resources:
            self.resources = {}
        else:
            self.resources = resources

        self.director_version = director_version

    def _parse(self, reader: EndiannessAwareStream, size: int):
        super()._parse(reader, size)

        self.director_version = self._parser.director_version

        for entry, resource in self._parser.entries:
            if entry.tag not in ('free', 'junk', '\x00\x00\x00\x00'):
                self.resources[entry.index] = resource

    def _serialize(self, writer: EndiannessAwareStream) -> None:
        serializer = DirectorArchiveSerializer(writer.endianness, self.director_version)
        serializer.serialize(writer.fp, self)


def load_director_archive(fp: BinaryIO):
    return DirectorArchiveResource().load(fp)



from setuptools import find_packages, setup

setup(
    name='directorfile',
    version='0.0.1',
    author='Adam Rimon',
    author_email='',
    description='A python module to read and manipulate Macromedia/Adobe Director files',
    url='https://github.com/Prilkop/directorfile',
    license='MIT License',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',

        'Intended Audience :: Developers',

        'Programming Language :: Python',
        'Topic :: File Formats',
        'Topic :: Utilities',

        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
)



import os
from enum import Enum, auto
from io import SEEK_END
from struct import pack, unpack
from typing import BinaryIO

from directorfile.archive import ApplicationArchiveResource
from directorfile.common import Endianness, EndiannessAwareStream, ParsingError


class ProjectorFormat(Enum):
    WINDOWS = auto()
    MAC_OS_PEF = auto()


class Projector:
    SUPPORTED_TAGS = {'PJ01'}
    _FORMAT_DISPLAY = {
        ProjectorFormat.WINDOWS: 'Windows',
        ProjectorFormat.MAC_OS_PEF: 'Mac OS (PEF)'

    }

    executable: bytes
    application: ApplicationArchiveResource

    _filename: str
    _format: ProjectorFormat

    def __init__(self, filename: str = ''):
        self._filename = filename

    def __repr__(self):
        if self._filename:
            return f'<Projector "{self._filename}">'
        else:
            return f'<Projector at {hex(id(self))}>'

    def load(self, fp: BinaryIO):
        if hasattr(fp, 'name'):
            self._filename = os.path.abspath(fp.name)

        position = self._locate_application(fp)
        fp.seek(0)
        self.executable = fp.read(position)
        self.application = ApplicationArchiveResource().load(fp, position)

        return self

    def _locate_application(self, fp: BinaryIO):
        fp.seek(0)
        head = fp.read(0x20)
        if head[0:2] == b'MZ':
            self._format = ProjectorFormat.WINDOWS
            reader = EndiannessAwareStream(fp, Endianness.LITTLE_ENDIAN)
            fp.seek(-4, SEEK_END)
            pj_position, = unpack('<I', fp.read(4))
            if pj_position <= 0x1000 or pj_position >= fp.tell() - 0x10:
                raise ParsingError('Not a projector file')
        elif head[10:18] == b'Joy!peff':
            self._format = ProjectorFormat.MAC_OS_PEF
            reader = EndiannessAwareStream(fp, Endianness.BIG_ENDIAN)
            pj_position = 0
        else:
            raise ParsingError('Unsupported executable format')

        self._pj_position = pj_position

        reader.jump(pj_position)
        tag = reader.read_tag()
        if tag in Projector.SUPPORTED_TAGS:
            container_position = reader.read_ui32()
        else:
            raise ParsingError(f'Unsupported PJ section: {tag}')
        return container_position

    def save(self, fp: BinaryIO, endianness: Endianness):
        fp.write(self.executable)
        self.application.save(fp, endianness)

        if self._format == ProjectorFormat.WINDOWS:
            fp.write(pack('<I', self._pj_position))


def load_projector(fp: BinaryIO, name: str = ''):
    return Projector(name).load(fp)
